// file kernel/c/h/numerix.h: C extensible integer definitions
/*-----------------------------------------------------------------------+
 |  Copyright 2005-2006, Michel Quercia (michel.quercia@prepas.org)      |
 |                                                                       |
 |  This file is part of Numerix. Numerix is free software; you can      |
 |  redistribute it and/or modify it under the terms of the GNU Lesser   |
 |  General Public License as published by the Free Software Foundation; |
 |  either version 2.1 of the License, or (at your option) any later     |
 |  version.                                                             |
 |                                                                       |
 |  The Numerix Library is distributed in the hope that it will be       |
 |  useful, but WITHOUT ANY WARRANTY; without even the implied warranty  |
 |  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  |
 |  Lesser General Public License for more details.                      |
 |                                                                       |
 |  You should have received a copy of the GNU Lesser General Public     |
 |  License along with the GNU MP Library; see the file COPYING. If not, |
 |  write to the Free Software Foundation, Inc., 59 Temple Place -       |
 |  Suite 330, Boston, MA 02111-1307, USA.                               |
 +-----------------------------------------------------------------------+
 |                                                                       |
 |                    Dfinitions pour l'interface C                     |
 |                                                                       |
 +-----------------------------------------------------------------------*/

/* Numerix C-library, public interface

   C extensible integers are defined as pointers to structures containing
   two fields of long type ...

   THIS IS NOT THE REALITY

   ... but it will be enough to make the C compiler happy without having
   to include all the low-level Numerix headers.

   Please,

     1. never make an xint of your own by directing some pointer towards
        a double-long structure. Use only the xx-functions described below
        to build xint values.

     2. never use an xint created for one mode (clong/dlong/slong) with
        xx-functions dealing with another mode.

     3. never use pointer arithmetic on xints.

*/

typedef struct xint_item {long lmax; long hd;} *xint; /* white lie */

/*
  check if one and only one mode is defined and setup the xx macro
  accordingly
*/

#if   defined(use_clong) + defined(use_dlong) + defined(use_slong) < 1
#error  "missing use_xlong definition"
#elif defined(use_clong) + defined(use_dlong) + defined(use_slong) > 1
#error  "multiple use_xlong definitions"
#endif

#if defined(use_clong)
#define xx(nom)  cx_##nom
#elif defined(use_dlong)
#define xx(nom)  dx_##nom
#elif defined(use_slong)
#define xx(nom)  sx_##nom
#endif

/* creation/destruction */
extern inline xint xx(new)() {return(NULL);}
extern inline void xx(free)(xint *_x) {if (_x) {free(*_x); *_x = NULL;}}

// file kernel/x/h/add.h: addition/subtraction of extensible integers
                              /* +------------+
                                 |  Addition  |
                                 +------------+ */
/*  
   entre :
   a,b = entiers extensibles
   _c = NULL ou pointeur sur un entier extensible
   s = 0 ou 1

   sortie :
   c <- a + (-1)^s*b
   si _c != NULL, *_c <- c
   retourne c
   */
xint xx(private_add)(xint *_c, xint a, xint b, long s);
                           /* +------------------+
                              |  Addition mixte  |
                              +------------------+ */
/*
  entre :
  a = entier extensible
  b = long ou Caml/Ocaml int
   _c = NULL ou pointeur sur un entier extensible
  s = 0 ou 1

  sortie :
  c <- a + (-1)^s*b
  si _c != NULL, *_c <- c
  retourne c
*/
xint xx(private_add_1)(xint *_c, xint a, long b, long s);
extern inline xint xx(add) (xint *_c, xint a, xint b) {return xx(private_add) (_c,a,b,0);}
extern inline xint xx(sub) (xint *_c, xint a, xint b) {return xx(private_add) (_c,a,b,1);}
extern inline xint xx(add_1)(xint *_c, xint a, long b) {return xx(private_add_1)(_c,a,b,0);}
extern inline xint xx(sub_1)(xint *_c, xint a, long b) {return xx(private_add_1)(_c,a,b,1);}
extern inline xint xx(f_add) (xint a, xint b) {return xx(private_add) (NULL,a,b,0);}
extern inline xint xx(f_sub) (xint a, xint b) {return xx(private_add) (NULL,a,b,1);}
extern inline xint xx(f_add_1) (xint a, long b) {return xx(private_add_1)(NULL,a,b,0);}
extern inline xint xx(f_sub_1) (xint a, long b) {return xx(private_add_1)(NULL,a,b,1);}
// file kernel/x/h/cmp.h: comparison of extensible integers
/*
  entre :
  a,b = entiers extensibles

  sortie :
  1 si a > b, 0 si a = b, -1 si a < b
*/
long xx(cmp)(xint a, xint b);
/*
  entre :
  a = entier extensible
  b = entier Caml/Ocaml ou long

  sortie :
  1 si a > b, 0 si a = b, -1 si a < b
*/
long xx(cmp_1)(xint a, long b);
/* versions spcialises */
extern inline long xx(sgn)(xint a){return((a->hd > 0) ? 1 : (a->hd < 0) ? -1 : 0);}
extern inline long xx(eq) (xint a,xint b) {return((xx(cmp) (a,b) == 0));}
extern inline long xx(neq) (xint a,xint b) {return((xx(cmp) (a,b) != 0));}
extern inline long xx(inf) (xint a,xint b) {return((xx(cmp) (a,b) < 0));}
extern inline long xx(infeq) (xint a,xint b) {return((xx(cmp) (a,b) <= 0));}
extern inline long xx(sup) (xint a,xint b) {return((xx(cmp) (a,b) > 0));}
extern inline long xx(supeq) (xint a,xint b) {return((xx(cmp) (a,b) >= 0));}
extern inline long xx(eq_1) (xint a,long b) {return((xx(cmp_1)(a,b) == 0));}
extern inline long xx(neq_1) (xint a,long b) {return((xx(cmp_1)(a,b) != 0));}
extern inline long xx(inf_1) (xint a,long b) {return((xx(cmp_1)(a,b) < 0));}
extern inline long xx(infeq_1)(xint a,long b) {return((xx(cmp_1)(a,b) <= 0));}
extern inline long xx(sup_1) (xint a,long b) {return((xx(cmp_1)(a,b) > 0));}
extern inline long xx(supeq_1)(xint a,long b) {return((xx(cmp_1)(a,b) >= 0));}
// file kernel/x/h/convert.h: extensible integer <-> long conversion
                            /* +---------------+
                               |  int -> xint  |
                               +---------------+ */
/*
   entre :
   a = long ou Caml/Ocaml int
   _b = NULL ou pointeur sur un entier extensible

   sortie :
   b <- a
   si _b != NULL, *_b <- b
   retourne b
*/
xint xx(copy_int)(xint *_b, long a);
extern inline xint xx(of_int)(long a) {return(xx(copy_int)(NULL,a));}
                            /* +---------------+
                               |  xint -> int  |
                               +---------------+ */
/*                                
   entre :
   a = entier extensible

   sortie :
   retourne le long ou Caml/Ocaml int de mme valeur que a

   erreur :
   INTEGER_OVERFLOW si |a| >= 2^30
*/
long xx(int_of)(xint a);
                         /* +---------------------+
                            |  Structure binaire  |
                            +---------------------+ */
/*
   entre :
   a = entier extensible

   sortie :
   si a <> 0 -> 1 + ceil(log_2(|a|))
   si a =  0 -> 0
*/
long xx(nbits)(xint a);
/*
  entre :
  a = entier extensible

  sortie :
  bits 0..30 de |a|
*/
long xx(lowbits)(xint a);
/*
  entre :
  a = entier extensible

  sortie :
  bits l-31..l-1 de |a|, l = nbits(a)
*/
long xx(highbits)(xint a);
/*
  entre :
  a = entier extensible
  n = indice

  sortie :
  bits 16*(n-1)..16*n-1 de |a|
*/
long xx(nth_word)(xint a, long n);
/*
  entre :
  a = entier extensible
  n = indice

  sortie :
  bit n de |a|
*/
long xx(nth_bit)(xint a, long n);
// file kernel/x/h/copy.h: extensible integer copying
                   /* +---------------------------------+
                      |  copie, valeur absolue, oppos  |
                      +---------------------------------+ */
/*
   entre :
   a = entier extensible
   _b = NULL ou pointeur sur un entier extensible
   s = 0 ou 1 ou 2

   sortie :
   si s = 0, b <- a
   si s = 1, b <- |a|
   si s = 2, b <- -a
   si _b != NULL, *_b <- b
   retourne b
   */
xint xx(private_copy)(xint *_b, xint a, long s);
extern inline xint xx(copy) (xint *_b, xint a) {return xx(private_copy)(_b,a,0);}
extern inline xint xx(abs) (xint *_b, xint a) {return xx(private_copy)(_b,a,1);}
extern inline xint xx(neg) (xint *_b, xint a) {return xx(private_copy)(_b,a,2);}
extern inline xint xx(f_abs) (xint a) {return xx(private_copy)(NULL,a,1);}
extern inline xint xx(f_neg) (xint a) {return xx(private_copy)(NULL,a,2);}
extern inline xint xx(f_copy)(xint a) {return xx(private_copy)(NULL,a,0);}
// file kernel/x/h/div.h: division of extensible integers
                         /* +---------------------+
                            |  Division gnrale  |
                            +---------------------+ */
/*
  entre :
  a,b = entiers extensibles
  _c  = NULL ou adresse d'un entier extensible
  _d  = NULL ou adresse d'un entier extensible
  mode = long

  contraintes :
  b != 0
  si _c et _d ne sont pas gaux  NULL alors ils sont distincts

  sortie :
  si mode & 3 = 0 : c <- floor(a/b),     d <- a - b*c (division tronque    )
  si mode & 3 = 1 : c <- floor(a/b+1/2), d <- a - b*c (division centre up  )
  si mode & 3 = 2 : c <- ceil(a/b),      d <- a - b*c (division majore     )
  si mode & 3 = 3 : c <- ceil(a/b-1/2),  d <- a - b*c (division centre down)
  si _c != NULL : *_c <- c
  si _d != NULL : *_d <- d

  valeur de retour  : c_api   ml_api
  si mode & 12 =  0 :  NULL     unit
  si mode & 12 =  4 :     c        c
  si mode & 12 =  8 :     d        d
  si mode & 12 = 12 :  NULL    (c,d)

  erreur :
  ZERO_DIVISOR si b = 0
  MULTIPLE_RESULT si _c == _d != NULL
*/
xint xx(private_quomod)(xint *_c, xint *_d, xint a, xint b, long mode);
extern inline void xx(gquomod)(xint *_c, xint *_d, xint a, xint b, long mode) { xx(private_quomod)(_c, _d, a,b, 0|(mode & 3));}
extern inline xint xx(gquo) (xint *_c, xint a, xint b, long mode) {return xx(private_quomod)(_c, NULL, a,b, 4|(mode & 3));}
extern inline xint xx(gmod) ( xint *_d, xint a, xint b, long mode) {return xx(private_quomod)(NULL,_d, a,b, 8|(mode & 3));}
extern inline xint xx(f_gquo) ( xint a, xint b, long mode) {return xx(private_quomod)(NULL,NULL, a,b, 4|(mode & 3));}
extern inline xint xx(f_gmod) ( xint a, xint b, long mode) {return xx(private_quomod)(NULL,NULL, a,b, 8|(mode & 3));}
extern inline void xx(quomod) (xint *_c, xint *_d, xint a, xint b ) { xx(private_quomod)(_c, _d, a,b, 0|0);}
extern inline xint xx(quo) (xint *_c, xint a, xint b ) {return xx(private_quomod)(_c, NULL, a,b, 4|0);}
extern inline xint xx(mod) ( xint *_d, xint a, xint b ) {return xx(private_quomod)(NULL,_d, a,b, 8|0);}
extern inline xint xx(f_quo) ( xint a, xint b ) {return xx(private_quomod)(NULL,NULL, a,b, 4|0);}
extern inline xint xx(f_mod) ( xint a, xint b ) {return xx(private_quomod)(NULL,NULL, a,b, 8|0);}
                      /* +--------------------------+
                         |  Division par un entier  |
                         +--------------------------+ */
/*
   entre :
   a  = entier extensible
   b  = long ou Caml/Ocaml int
   _c = NULL ou adresse d'un entier extensible
   mode = long

   sortie :
   si mode & 3 = 0 : c <- floor(a/b),     d <- a - b*c (division tronque)
   si mode & 3 = 1 : c <- floor(a/b+1/2), d <- a - b*c (division centre up)
   si mode & 3 = 2 : c <- ceil(a/b),      d <- a - b*c (division majore)
   si mode & 3 = 3 : c <- ceil(a/b-1/2),  d <- a - b*c (division centre down)
   si _c != NULL : *_c <- c

   valeur de retour  : c_api   ml_api
   si mode & 12 =  0 :     d     unit
   si mode & 12 =  4 :     c        c
   si mode & 12 =  8 :     d        d
   si mode & 12 = 12 :     c    (c,d)

   erreur :
   ZERO_DIVISOR si b = 0
*/
long xx(private_quomod_1)(xint *_c, xint a, long b, long mode);
extern inline long xx(gquomod_1)(xint *_c, xint a, long b, long mode) {return xx(private_quomod_1)(_c, a,b, 8|(mode & 3));}
extern inline xint xx(gquo_1) (xint *_c, xint a, long b, long mode) {return (xint)xx(private_quomod_1)(_c, a,b, 4|(mode & 3));}
extern inline long xx(gmod_1) ( xint a, long b, long mode) {return xx(private_quomod_1)(NULL, a,b, 8|(mode & 3));}
extern inline xint xx(f_gquo_1) ( xint a, long b, long mode) {return (xint)xx(private_quomod_1)(NULL, a,b, 4|(mode & 3));}
extern inline long xx(f_gmod_1) ( xint a, long b, long mode) {return xx(private_quomod_1)(NULL, a,b, 8|(mode & 3));}
extern inline long xx(quomod_1) (xint *_c, xint a, long b ) {return xx(private_quomod_1)(_c, a,b, 8|0);}
extern inline xint xx(quo_1) (xint *_c, xint a, long b ) {return (xint)xx(private_quomod_1)(_c, a,b, 4|0);}
extern inline long xx(mod_1) ( xint a, long b ) {return xx(private_quomod_1)(NULL, a,b, 8|0);}
extern inline xint xx(f_quo_1) ( xint a, long b ) {return (xint)xx(private_quomod_1)(NULL, a,b, 4|0);}
extern inline long xx(f_mod_1) ( xint a, long b ) {return xx(private_quomod_1)(NULL, a,b, 8|0);}
// file kernel/x/h/dump.h: debugging facilities
/* ---------------------------------------- Affichage hexadcimal
  entre :
  a = entier extensible

  sortie :
  affiche a
*/
void xx(dump)(xint a);
/* ---------------------------------------- Erreur interne
   entre :
   msg = chane de caractres
   n   = nombre d'entiers  afficher
   ... = suite d'entiers extensibles

   sortie :
   affiche le message et les entiers donns, puis termine le programme
*/
void xx(internal_error)(char *msg, int n, ...) __attribute__((noreturn));
// file kernel/x/h/fact.h: factorial
/*
  entre :
  _a = NULL ou pointeur sur un entier extensible
  n = long ou Caml/Ocaml int >= 0

  sortie :
  a <- n!
  si _a != NULL, *_a <- a
  retourne a

  erreur :
  NEGATIVE_BASE si n < 0
*/
xint xx(fact)(xint *_a, long n);
extern inline xint xx(f_fact)(long n) {return xx(fact)(NULL,n);}
// file kernel/x/h/gcd.h: Greatest common divisor
              /* +--------------------------------------+
                 |  Dveloppement en fraction continue  |
                 +--------------------------------------+ */
/*
  entre :
  a,b = entiers extensibles
  _d,_u,_v,_p,_q = NULL ou pointeurs vers des entiers extensibles
  mode = long

  sortie :
  calcule d,u,v,p,q tels que :

             u*a - v*b = d = pgcd positif(a,b)
             p*b = q*a = ppcm(a,b) avec le signe de a*b
             p*u - q*v = 1

  pour x = d,u,v,p,q : si _x != NULL, *_x <- x

  valeur de retour : c_api       ml_api
  si mode & 3 = 0  :  NULL         unit
  si mode & 3 = 1  :     d           d
  si mode & 3 = 2  :  NULL      (d,u,v)
  si mode & 3 = 3  :  NULL  (d,u,v,p,q)

  erreur :
  MULTIPLE_RESULT si (_d,_u,_v,_p,_q) contient deux pointeurs non nuls gaux
*/
xint xx(private_cfrac)(xint *_d, xint *_u, xint *_v, xint *_p, xint *_q, xint a, xint b, long mode);
extern inline xint xx(gcd)(xint *_d, xint a, xint b) {
  return xx(private_cfrac)(_d,NULL,NULL,NULL,NULL,a,b,1);
}
extern inline void xx(gcd_ex)(xint *_d, xint *_u, xint *_v, xint a, xint b) {
    xx(private_cfrac)(_d,_u,_v,NULL,NULL,a,b,0);
}
extern inline void xx(cfrac)(xint *_d, xint *_u, xint *_v, xint *_p, xint *_q, xint a, xint b) {
    xx(private_cfrac)(_d,_u,_v,_p,_q,a,b,0);
}
extern inline xint xx(f_gcd)(xint a, xint b) {
    return xx(private_cfrac)(NULL,NULL,NULL,NULL,NULL,a,b,1);
}
// file kernel/x/h/mul.h: multiplication/square of extensible integers
                          /* +------------------+
                             |  Multiplication  |
                             +------------------+ */
/*
   entre :
   a,b = entiers extensibles
   _c = NULL ou pointeur sur un entier extensible

   sortie :
   c <- a * b
   si _c != NULL, *_c <- c
   retourne c
*/
xint xx(mul)(xint *_c, xint a, xint b);
                    /* +------------------------------+
                       |  Multiplication par un long  |
                       +------------------------------+ */
/*
   entre :
   a   = entier extensible
   b   = long ou Caml/Ocaml int
   _c = NULL ou pointeur sur un entier extensible

   sortie :
   c <- a * b
  si _c != NULL, *_c <- c
  retourne c
*/
xint xx(mul_1)(xint *_c, xint a, long b);
                               /* +---------+
                                  |  Carr  |
                                  +---------+ */
/*
   entre :
   a   = entier extensible
   _b = NULL ou pointeur sur un entier extensible

   sortie :
   b <- a^2
   si _b != NULL, *_b <- b
   retourne b
*/
xint xx(sqr)(xint *_b, xint a);
extern inline xint xx(f_mul) (xint a, xint b) {return xx(mul) (NULL,a,b);}
extern inline xint xx(f_mul_1)(xint a, long b) {return xx(mul_1)(NULL,a,b);}
extern inline xint xx(f_sqr) (xint a) {return xx(sqr) (NULL,a); }
// file kernel/x/h/pow.h: exponentiation
/*
  entre :
  a  = entier extensible
  _b = NULL ou pointeur sur un entier extensible
  p  = exposant >= 0

  sortie :
  b <- a^p
  si _b != NULL, *_b <- b
  retourne b

  erreur :
  NEGATIVE_EXPONENT si p < 0
*/
xint xx(pow)(xint *_b, xint a, long p);
/*
  entre :
  a  = long ou Caml/Ocaml int
  _b = NULL ou pointeur sur un entier extensible
  p  = exposant >= 0

  sortie :
  b <- a^p
  si _b != NULL, *_b <- b
  retourne b

  erreur :
  NEGATIVE_EXPONENT si p < 0
*/
xint xx(pow_1)(xint *_b, long a, long p);
extern inline xint xx(f_pow) (xint a, long p) {return xx(pow) (NULL,a,p);}
extern inline xint xx(f_pow_1)(long a, long p) {return xx(pow_1)(NULL,a,p);}
// file kernel/x/h/powmod.h: modular exponentiation
/*
  entre :
  a,b,c = entiers extensibles
  _d = NULL ou pointeur sur un entier extensible
  mode = long

  sortie :
  d <- gmod(a^b, c, mode)
  si _d != NULL, *_d <- d
  retourne d

  erreur :
  NEGATIVE_EXPONENT si b < 0
  ZERO_DIVISOR      si c = 0
*/
xint xx(private_powmod)(xint *_d, xint a, xint b, xint c, long mode);
extern inline xint xx(gpowmod)(xint *_d, xint a, xint b, xint c, long mode) {
    return xx(private_powmod)(_d,a,b,c,mode);
}
extern inline xint xx(powmod)(xint *_d, xint a, xint b, xint c) {
    return xx(private_powmod)(_d,a,b,c,0);
}
extern inline xint xx(f_gpowmod)(xint a, xint b, xint c, long mode) {
    return xx(private_powmod)(NULL,a,b,c,mode);
}
extern inline xint xx(f_powmod)(xint a, xint b, xint c) {
    return xx(private_powmod)(NULL,a,b,c,0);
}
// file kernel/x/h/prime.h: primality test
/*
  entre :
  a = entier extensible (isprime) ou entier long (isprime_1)

  sortie = 0 : a est compos,
  sortie = 1 : a passe le test mais |a| est trop grand pour qu'on soit sr
               que a est premier.
  sortie = 2 : a est premier,
*/
long xx(isprime)(xint a);
long xx(isprime_1)(long a);
// file kernel/x/h/random.h: random extensible integers
/* initialisation du gnrateur alatoire
   si n != 0, le prend comme graine
   si n = 0,  prend l'heure en secondes comme graine */
void xx(random_init)(long n);
/* Nombre alatoire
   entre :
   _a = NULL ou pointeur vers un entier extensible
   n = longueur >= 0
   mode = long C

   sortie :
   a <- entier alatoire entre -2^n-1 et 2^n-1
   les bits 0 et 1 de mode restreignent la plage alatoire de la manire suivante :
   b0=0 => le signe est forc >= 0
   b1=1 => le bit de rang n-1 est forc = 1 si n > 0

   si _a != NULL, *_a <- a
   retourne a

   erreur :
   NEGATIVE_SIZE si n < 0
*/
xint xx(private_random)(xint *_a, long n, long mode);
/*
  nrandom(n)  -> [0, 2^n[
  zrandom(n)  -> ]-2^n, 2^n[
  nrandom1(n) -> [2^(n-1), 2^n[
  zrandom1(n) -> ]-2^n, -2^(n-1)] U [2^(n-1), 2^n[
*/
extern inline xint xx(nrandom) (xint *_a, long n) {return xx(private_random)(_a,n,0);}
extern inline xint xx(zrandom) (xint *_a, long n) {return xx(private_random)(_a,n,1);}
extern inline xint xx(nrandom1)(xint *_a, long n) {return xx(private_random)(_a,n,2);}
extern inline xint xx(zrandom1)(xint *_a, long n) {return xx(private_random)(_a,n,3);}
extern inline xint xx(f_nrandom) (long n) {return xx(private_random)(NULL,n,0);}
extern inline xint xx(f_zrandom) (long n) {return xx(private_random)(NULL,n,1);}
extern inline xint xx(f_nrandom1) (long n) {return xx(private_random)(NULL,n,2);}
extern inline xint xx(f_zrandom1) (long n) {return xx(private_random)(NULL,n,3);}
// file kernel/x/h/root.h: pth-root of extensible integers
/*
  entre :
  a   = entier extensible
   _b = NULL ou pointeur sur un entier extensible
  p   = entier > 0
  mode = 0,1 ou 2

  sortie :
  si mode & 3 = 0: b <- floor(a^(1/p))
  si mode & 3 = 1: b <- floor(a^(1/p)+1/2)
  si mode & 3 = 2: b <- ceil(a^(1/p))
  si mode & 3 = 1: b <- ceil(a^(1/p)-1/2) (= floor(a^(1/p)+1/2))
   si _b != NULL, *_b <- b
   retourne b

  erreurs :
  NEGATIVE_EXPONENT si p <= 0
  NEGATIVE_BASE si a < 0 et p est pair
*/
xint xx(private_root)(xint *_b, xint a, long p, long mode);
extern inline xint xx(root) (xint *_b, xint a, long p ) {return xx(private_root)(_b, a,p,0);}
extern inline xint xx(groot) (xint *_b, xint a, long p, long mode) {return xx(private_root)(_b, a,p,mode);}
extern inline xint xx(f_root) ( xint a, long p ) {return xx(private_root)(NULL,a,p,0);}
extern inline xint xx(f_groot)( xint a, long p, long mode) {return xx(private_root)(NULL,a,p,mode);}
// file kernel/x/h/shift.h: shift of extensible integers
                             /* +-------------+
                                |  Dcalages  |
                                +-------------+ */
/*
   entre :
   a  = entier extensible
   _b = NULL ou pointeur sur un entier extensible
   n  = entier non sign
   sens = 0 ou 1

   sortie :
   si sens = 0: b <- a >> n
   si sens = 1: b <- a << n
   si _b != NULL, *_b <- b
   retourne b
*/
xint xx(private_shift)(xint *_b, xint a, unsigned long n, long sens);
extern inline xint xx(shl)(xint *_b, xint a, long n) {
    return((n >= 0) ? xx(private_shift)(_b,a, n,1)
                    : xx(private_shift)(_b,a,-n,0));
}
extern inline xint xx(shr)(xint *_b, xint a, long n) {
    return((n < 0) ? xx(private_shift)(_b,a,-n,1)
                    : xx(private_shift)(_b,a, n,0));
}
extern inline xint xx(f_shl)(xint a, long n) {
    return((n >= 0) ? xx(private_shift)(NULL,a, n,1)
                    : xx(private_shift)(NULL,a,-n,0));
}
extern inline xint xx(f_shr)(xint a, long n) {
    return((n < 0) ? xx(private_shift)(NULL,a,-n,1)
                    : xx(private_shift)(NULL,a, n,0));
}
                             /* +-------------+
                                |  Dcoupage  |
                                +-------------+ */
/*
   entre :
   a = entier extensible
   _b,_c = NULL ou pointeurs vers des entiers extensibles
   n = longueur >= 0

   contraintes :
   en mode Caml/Ocaml, _b et _c ont la mme validit (NULL/non NULL)
   en mode C, les validits de _b et _c sont indpendantes
   lorsque _b et _c sont des pointeurs valides, ils sont distincts

   sortie :
   b <- sgn(a)*floor(|a|/2^n)
   c <- sgn(a)*(|a| mod 2^n)
   si_ b != NULL, *_b <- b 
   si _c != NULL, *_c <- c
   si _b = _c = NULL retourne le couple (b,c) (Caml/Ocaml uniquement)
  
   erreur :
   NEGATIVE_INDEX si n < 0
   MULTIPLE_RESULT si _b == _c != NULL
*/
void xx(split)(xint *_b, xint *_c, xint a, long n);
                           /* +-----------------+
                              |  Concatnation  |
                              +-----------------+ */
/*
   entre :
   a,b = entiers extensibles
   _c = NULL ou pointeur sur un entier extensible
   n   = longueur >= 0

   sortie :
   c <- a + b*2^n
   si _c != NULL, *_c <- c
   retourne c

   erreur :
   NEGATIVE_INDEX si n < 0
*/
xint xx(join)(xint *_c, xint a, xint b, long n);
extern inline xint xx(f_join)(xint a, xint b, long n) {
    return xx(join)(NULL,a,b,n);
}
// file kernel/x/h/sqrt.h: square root of extensible integers
/*
  entre :
  a   = entier extensible
   _b = NULL ou pointeur sur un entier extensible
  mode = long

  sortie :
  si mode & 3 = 0: b <- floor(sqrt(a))
  si mode & 3 = 1: b <- floor(sqrt(a)+1/2)
  si mode & 3 = 2: b <- ceil(sqrt(a))
  si mode & 3 = 3: b <- ceil(sqrt(a)-1/2) (= floor(sqrt(a)+1/2))
  si _b != NULL, *_b <- b
  retourne b

  erreur :
  NEGATIVE_BASE si a < 0
*/
xint xx(private_sqrt)(xint *_b, xint a, long mode);
extern inline xint xx(sqrt) (xint *_b, xint a ) {return xx(private_sqrt)(_b, a,0);}
extern inline xint xx(gsqrt) (xint *_b, xint a, long mode) {return xx(private_sqrt)(_b, a,mode);}
extern inline xint xx(f_sqrt) ( xint a ) {return xx(private_sqrt)(NULL,a,0);}
extern inline xint xx(f_gsqrt)( xint a, long mode) {return xx(private_sqrt)(NULL,a,mode);}
// file kernel/x/h/string.h: conversion from and to strings
                         /* +--------------------+
                            |  chane -> entier  |
                            +--------------------+ */
/*
  entre :
  _a = NULL ou pointeur sur entier extensible
  s   = chane de caractres

  sortie :
  a <- valeur de s
  si _a != NULL, *_a <- a
  retourne a

  erreur :
  INVALID_STRING si s contient un caractre non reconnu ou s est vide
  ou rduite au signe et au prfixe

  remarques :
  les signes + et - en dbut de chane sont accepts
  les prfixes 0x, 0X, 0o, 0O, 0b, 0B aprs le signe sont reconnus
  pour une chane hexadcimale, les lettres a..f et A..F sont acceptes
*/
xint xx(copy_string)(xint *_a, char *s);
extern inline xint xx(of_string)(char *s) {return xx(copy_string)(NULL,s);}
                     /* +-----------------------------+
                        |  entier -> chane dcimale  |
                        +-----------------------------+ */
/* 
   entre :
   a = entier extensible

   sortie :
   retourne la reprsentation dcimale de a

   remarques :
   la chane doit tre libre par appel  free (c_api)
   la chane cre ne contient pas de zro non significatif
   le signe n'est indiqu que si a < 0
*/
char *xx(string_of)(xint a);
                       /* +-------------------------+
                          |  entier -> chane hexa  |
                          +-------------------------+ */
/* 
   entre :
   a = entier extensible

   sortie :
   retourne la reprsentation hexadcimale de a

   remarques :
   la chane doit tre libre par appel  free (c_api)
   les chiffres >= 10 sont cods A..F
   la chane cre ne contient pas de zro non significatif
   le signe n'est indiqu que si a < 0
*/
char *xx(hstring_of)(xint a);
                      /* +---------------------------+
                         |  entier -> chane octale  |
                         +---------------------------+ */
/* 
   entre :
   a = entier extensible

   sortie :
   retourne la reprsentation octale de a

   remarques :
   la chane doit tre libre par appel  free (c_api)
   la chane cre ne contient pas de zro non significatif
   le signe n'est indiqu que si a < 0
*/
char *xx(ostring_of)(xint a);
                     /* +----------------------------+
                        |  entier -> chane binaire  |
                        +----------------------------+ */
/* 
   entre :
   a = entier extensible

   sortie :
   retourne la reprsentation binaire de a

   remarques :
   la chane doit tre libre par appel  free (c_api)
   la chane cre ne contient pas de zro non significatif
   le signe n'est indiqu que si a < 0
*/
char *xx(bstring_of)(xint a);
// file kernel/x/h/chrono.h: timing facility
/* affiche sur stdout le temps utilisateur,
                      la diffrence avec le temps prcdent,
                      le message donn en argument
*/
void chrono(char *msg);
